@@ -7,7 +7,7 @@ import cgi
 
 try:
     from textwrap import wrap
-except ImportError:
+except ImportError:  # pragma: no cover
     pass
 
 from html2text.compat import urlparse, HTMLParser
@@ -29,7 +29,7 @@ from html2text.utils import (
     skipwrap
 )
 
-__version__ = (2015, 6, 12)
+__version__ = (2015, 11, 4)
 
 
 # TODO:
@@ -50,31 +50,36 @@ class HTML2Text(HTMLParser.HTMLParser):
         self.split_next_td = False
         self.td_count = 0
         self.table_start = False
-        self.unicode_snob = config.UNICODE_SNOB
-        self.escape_snob = config.ESCAPE_SNOB
+        self.unicode_snob = config.UNICODE_SNOB  # covered in cli
+        self.escape_snob = config.ESCAPE_SNOB  # covered in cli
         self.links_each_paragraph = config.LINKS_EACH_PARAGRAPH
-        self.body_width = bodywidth
-        self.skip_internal_links = config.SKIP_INTERNAL_LINKS
-        self.inline_links = config.INLINE_LINKS
-        self.protect_links = config.PROTECT_LINKS
-        self.google_list_indent = config.GOOGLE_LIST_INDENT
-        self.ignore_links = config.IGNORE_ANCHORS
-        self.ignore_images = config.IGNORE_IMAGES
-        self.images_to_alt = config.IMAGES_TO_ALT
-        self.images_with_size = config.IMAGES_WITH_SIZE
-        self.ignore_emphasis = config.IGNORE_EMPHASIS
-        self.bypass_tables = config.BYPASS_TABLES
-        self.google_doc = False
-        self.ul_item_mark = '*'
-        self.emphasis_mark = '_'
+        self.body_width = bodywidth  # covered in cli
+        self.skip_internal_links = config.SKIP_INTERNAL_LINKS  # covered in cli
+        self.inline_links = config.INLINE_LINKS  # covered in cli
+        self.protect_links = config.PROTECT_LINKS  # covered in cli
+        self.google_list_indent = config.GOOGLE_LIST_INDENT  # covered in cli
+        self.ignore_links = config.IGNORE_ANCHORS  # covered in cli
+        self.ignore_images = config.IGNORE_IMAGES  # covered in cli
+        self.images_to_alt = config.IMAGES_TO_ALT  # covered in cli
+        self.images_with_size = config.IMAGES_WITH_SIZE  # covered in cli
+        self.ignore_emphasis = config.IGNORE_EMPHASIS  # covered in cli
+        self.bypass_tables = config.BYPASS_TABLES  # covered in cli
+        self.google_doc = False  # covered in cli
+        self.ul_item_mark = '*'  # covered in cli
+        self.emphasis_mark = '_'  # covered in cli
         self.strong_mark = '**'
+        self.single_line_break = config.SINGLE_LINE_BREAK  # covered in cli
+        self.use_automatic_links = config.USE_AUTOMATIC_LINKS  # covered in cli
+        self.hide_strikethrough = False  # covered in cli
+        self.mark_code = config.MARK_CODE
         self.single_line_break = config.SINGLE_LINE_BREAK
         self.use_automatic_links = config.USE_AUTOMATIC_LINKS
-        self.mark_code = config.MARK_CODE
+        self.wrap_links = config.WRAP_LINKS  # covered in cli
+        self.tag_callback = None
 
-        if out is None:
+        if out is None:  # pragma: no cover
             self.out = self.outtextf
-        else:
+        else:  # pragma: no cover
             self.out = out
 
         # empty list to store output characters before they are "joined"
@@ -167,13 +172,14 @@ class HTML2Text(HTMLParser.HTMLParser):
         charref = self.charref(c)
         if not self.code and not self.pre:
             charref = cgi.escape(charref)
-        self.o(charref, 1)
+        self.handle_data(charref, True)
 
     def handle_entityref(self, c):
         entityref = self.entityref(c)
-        if not self.code and not self.pre and entityref != '&nbsp_place_holder;':
+        if (not self.code and not self.pre
+                and entityref != '&nbsp_place_holder;'):
             entityref = cgi.escape(entityref)
-        self.o(entityref, 1)
+        self.handle_data(entityref, True)
 
     def handle_starttag(self, tag, attrs):
         self.handle_tag(tag, attrs, 1)
@@ -189,9 +195,8 @@ class HTML2Text(HTMLParser.HTMLParser):
         self.a list. If the set of attributes is not found, returns None
         :rtype: int
         """
-        if 'href' not in attrs:
+        if 'href' not in attrs:  # pragma: no cover
             return None
-
         i = -1
         for a in self.a:
             i += 1
@@ -277,9 +282,14 @@ class HTML2Text(HTMLParser.HTMLParser):
         else:
             attrs = dict(attrs)
 
+        if self.tag_callback is not None:
+            if self.tag_callback(self, tag, attrs, start) is True:
+                return
+
         # first thing inside the anchor tag is another tag that produces some output
-        if start and not self.maybe_automatic_link is None and \
-             tag not in ['p', 'div', 'style', 'dl', 'dt'] and (tag != "img" or self.ignore_images):
+        if (start and not self.maybe_automatic_link is None
+                and tag not in ['p', 'div', 'style', 'dl', 'dt']
+                and (tag != "img" or self.ignore_images)):
             self.o("[")
             self.maybe_automatic_link = None
             self.empty_link = False
@@ -405,7 +415,13 @@ class HTML2Text(HTMLParser.HTMLParser):
                             self.empty_link = False
                             self.maybe_automatic_link = None
                         if self.inline_links:
-                            self.o("](" + escape_md(a['href']) + ")")
+                            try:
+                                title = escape_md(a['title'])
+                            except KeyError:
+                                self.o("](" + escape_md(a['href']) + ")")
+                            else:
+                                self.o("](" + escape_md(a['href'])
+                                       + ' "' + title + '" )')
                         else:
                             i = self.previousIndex(a)
                             if i is not None:
@@ -568,13 +584,16 @@ class HTML2Text(HTMLParser.HTMLParser):
 
     # TODO: Add docstring for these one letter functions
     def pbr(self):
+        "Pretty print has a line break"
         if self.p_p == 0:
             self.p_p = 1
 
     def p(self):
+        "Set pretty print to 1 or 2 lines"
         self.p_p = 1 if self.single_line_break else 2
 
     def soft_br(self):
+        "Soft breaks"
         self.pbr()
         self.br_toggle = '  '
 
@@ -683,7 +702,7 @@ class HTML2Text(HTMLParser.HTMLParser):
             self.out(data)
             self.outcount += 1
 
-    def handle_data(self, data):
+    def handle_data(self, data, entity_char=False):
         if r'\/script>' in data:
             self.quiet -= 1
 
@@ -692,7 +711,8 @@ class HTML2Text(HTMLParser.HTMLParser):
 
         if not self.maybe_automatic_link is None:
             href = self.maybe_automatic_link
-            if href == data and self.absolute_url_matcher.match(href) and self.use_automatic_links:
+            if (href == data and self.absolute_url_matcher.match(href)
+                    and self.use_automatic_links):
                 self.o("<" + data + ">")
                 self.empty_link = False
                 return
@@ -701,11 +721,11 @@ class HTML2Text(HTMLParser.HTMLParser):
                 self.maybe_automatic_link = None
                 self.empty_link = False
 
-        if not self.code and not self.pre:
+        if not self.code and not self.pre and not entity_char:
             data = escape_md_section(data, snob=self.escape_snob)
         self.o(data, 1)
 
-    def unknown_decl(self, data):
+    def unknown_decl(self, data):  # pragma: no cover
         # TODO: what is this doing here?
         pass
 
@@ -719,9 +739,12 @@ class HTML2Text(HTMLParser.HTMLParser):
             return unifiable_n[c]
         else:
             try:
-                return unichr(c)
-            except NameError:  # Python3
-                return chr(c)
+                try:
+                    return unichr(c)
+                except NameError:  # Python3
+                    return chr(c)
+            except ValueError:  # invalid unicode
+                return ''
 
     def entityref(self, c):
         if not self.unicode_snob and c in config.UNIFIABLE.keys():
@@ -779,9 +802,14 @@ class HTML2Text(HTMLParser.HTMLParser):
         assert wrap, "Requires Python 2.3."
         result = ''
         newlines = 0
+        # I cannot think of a better solution for now.
+        # To avoid the non-wrap behaviour for entire paras
+        # because of the presence of a link in it
+        if not self.wrap_links:
+            self.inline_links = False
         for para in text.split("\n"):
             if len(para) > 0:
-                if not skipwrap(para):
+                if not skipwrap(para, self.wrap_links):
                     result += "\n".join(wrap(para, self.body_width))
                     if para.endswith('  '):
                         result += "  \n"
@@ -804,7 +832,9 @@ class HTML2Text(HTMLParser.HTMLParser):
         return result
 
 
-def html2text(html, baseurl='', bodywidth=config.BODY_WIDTH):
+def html2text(html, baseurl='', bodywidth=None):
+    if bodywidth is None:
+        bodywidth = config.BODY_WIDTH
     h = HTML2Text(baseurl=baseurl, bodywidth=bodywidth)
 
     return h.handle(html)
